import html
import pathlib
from typing import Optional

import jinja2
from fastapi import APIRouter, Depends, HTTPException
from starlette import status

from api import db, get_conn
from model.patch_request import APIPatchRequestAction, PatchRequestFormRequest, InformalPatchRequestFormRequest, \
    PatchRequestAction, PatchActionType, InsertType, InsertPort
from model.settings import settings
from model.wapi.cntl import APIToken, Mgr
from model.wapi.nd import Port, Module, Room, Bldg, Site
from util.auth import check_auth
from util.util import send_email, map_array_of_dicts
from util.wapi_util import execute_wapi_function

router = APIRouter(
    prefix="/api/patch_request",
    tags=["patch_request"]
)
js_mods = ['patchrequest']


@router.get('/sites')
async def get_available_sites(token: APIToken = Depends(check_auth), conn=Depends(get_conn)):
    user = Mgr.check_token(conn, token)
    available_fq_names = list(settings.patch_request_assignment.keys())
    ta = [{
        'idx': site,
        'name': 'nd.site.list',
        'old': {'fq_name': site}
    } for site in available_fq_names]
    r = execute_wapi_function(conn, ta, user=user.login_name)
    return [r[site][0] for site in available_fq_names]


@router.get('/insert_types')
async def get_insert_types(token: APIToken = Depends(check_auth), conn=Depends(get_conn)) -> list[InsertType]:
    Mgr.check_token(conn, token)

    module_types: dict[str, InsertType] = {
        '1xETH (Gbit)': InsertType(name='1xETH (Gbit)', ports=[
            InsertPort(type='ETH', type_group='ETH', count=1, proto='Ethernet', name_prefix='', name_suffix='::1'),
        ]),
        '2xETH (100 Mbit)': InsertType(name='2xETH (100 Mbit)', ports=[
            InsertPort(type='ETH', type_group='ETH', count=1, proto='Ethernet', name_prefix='', name_suffix='/1'),
            InsertPort(type='ETH', type_group='ETH', count=1, proto='Ethernet', name_prefix='', name_suffix='/2'),
        ]),
        '1xETH (100 Mbit) + 1xTEL': InsertType(name='1xETH (100 Mbit) + 1xTEL', ports=[
            InsertPort(type='ETH', type_group='ETH', count=1, proto='Ethernet', name_prefix='', name_suffix='/1'),
            InsertPort(type='TEL', type_group='TEL', count=1, proto='ISDN', name_prefix='', name_suffix='/A'),
        ]),
        '1xETH (100 Mbit) + 2xTEL': InsertType(name='1xETH (100 Mbit) + 2xTEL', ports=[
            InsertPort(type='ETH', type_group='ETH', count=1, proto='Ethernet', name_prefix='', name_suffix='/1'),
            InsertPort(type='TEL', type_group='TEL', count=1, proto='ISDN', name_prefix='', name_suffix='/A'),
            InsertPort(type='TEL', type_group='TEL', count=1, proto='ISDN', name_prefix='', name_suffix='/B'),
        ])
    }

    return list(module_types.values())


@router.post('/informal')
async def handle_informal_request(informal_patch_request_form_request: InformalPatchRequestFormRequest,
                                  token: APIToken = Depends(check_auth),
                                  conn=Depends(get_conn),
                                  ):
    user = Mgr.check_token(conn, token)

    r = execute_wapi_function(conn, [
        {'idx': 'site_list', 'name': 'nd.site.list', 'old': {'gpk': informal_patch_request_form_request.site.gpk}}
    ], user=user.login_name)

    if len(r['site_list']) == 0:
        raise HTTPException(
            status_code=status.HTTP_400_BAD_REQUEST,
            detail='Requested site could not be found!'
        )

    informal_patch_request_form_request.site = Site(**r['site_list'][0])

    effective_site_name = informal_patch_request_form_request.site.parent_fq_name
    if effective_site_name is None:
        effective_site_name = informal_patch_request_form_request.site.fq_name

    if effective_site_name not in settings.patch_request_assignment.keys():
        raise HTTPException(
            status_code=status.HTTP_400_BAD_REQUEST,
            detail='The request contained unprocessable compounds ({})! '
                   'Please write an email in order to request this patch.'.format(
                    informal_patch_request_form_request.site.parent_fq_name
                    or informal_patch_request_form_request.site.fq_name
                    )
        )

    receiver = settings.patch_request_assignment[effective_site_name]
    successful = send_informal_petch_request(informal_patch_request_form_request.comment,
                                             mgr=user,
                                             receiver=receiver,
                                             reply_to=informal_patch_request_form_request.reply_to)

    if successful:
        send_informal_patch_request_confirm(informal_patch_request_form_request.comment,
                                            informal_patch_request_form_request.reply_to or user.email)
        return {'result': 'success'}

    raise HTTPException(status_code=status.HTTP_500_INTERNAL_SERVER_ERROR)


@router.post('')
async def handle_request(patch_request_form_request: PatchRequestFormRequest,
                         token: APIToken = Depends(check_auth),
                         conn=Depends(get_conn),
                         ):
    user = Mgr.check_token(conn, token)

    actions = validate_patch_request_actions(conn, user, patch_request_form_request.actions)

    split = map_array_of_dicts(actions, lambda x: x.site.parent_fq_name or x.site.fq_name or 'Unassigned')

    unassigned_sites = list(filter(lambda x: x not in settings.patch_request_assignment.keys(), split.keys()))
    if len(unassigned_sites) > 0:
        raise HTTPException(
            status_code=status.HTTP_400_BAD_REQUEST,
            detail='The request contained unprocessable compounds ({})! '
                   'Please write an email in order to request this patch.'.format(', '.join(unassigned_sites))
        )

    all_successful = True

    for site_name, receiver in settings.patch_request_assignment.items():
        if len(split.get(site_name, [])) > 0:
            all_successful &= send_patch_request(split[site_name], user, receiver=receiver,
                                                 reply_to=patch_request_form_request.reply_to)

    if all_successful:
        send_patch_request_confirm(actions, patch_request_form_request.reply_to or user.email)
        return {'result': 'success'}

    raise HTTPException(status_code=status.HTTP_500_INTERNAL_SERVER_ERROR)


def validate_patch_request_actions(conn, user, actions: list[APIPatchRequestAction]) -> list[PatchRequestAction]:
    internal_actions = []
    for action in actions:
        internal_action = PatchRequestAction()
        # TODO handle unpatch vs missing parameters
        if action.bcd_name is not None:
            if not user.has_bcd(conn, action.bcd_name):
                bcd_name = action.bcd_name
                raise HTTPException(
                    status_code=status.HTTP_403_FORBIDDEN,
                    detail=f'You do not have permission to request patch for BCD {bcd_name}'
                )
            bcd_result = execute_wapi_function(conn, [
                {'idx': 'bcd_list', 'name': 'nd.bcd.list', 'old': {'name': action.bcd_name}},
                {'idx': 'vlan_list', 'name': 'nd.vlan.list', 'inner_join_ref': {'bcd_list': 'default'}},
            ], user=user.login_name)
            internal_action.bcd = Port(**bcd_result['bcd_list'][0])
            if len(bcd_result['vlan_list']) > 0:
                internal_action.vlan_ids = list(
                    map(lambda v: '{compound}: {id}'.format(compound=v.get('net_instnc', 'UNKNOWN?'), id=v['id']),
                        bcd_result['vlan_list']))
            else:
                internal_action.vlan_ids = ['VLAN does not exist in any compound.']

        elif action.action == PatchActionType.PATCH:
            raise HTTPException(
                status_code=status.HTTP_400_BAD_REQUEST,
                detail='BCD is required for patch action'
            )

        if action.action != PatchActionType.CHANGE_INSERT and len(action.new_insert_assignment) > 0:
            raise HTTPException(
                status_code=status.HTTP_400_BAD_REQUEST,
                detail='New insert assignments are only allowed for change_insert actions'
            )

        for insert_assignment in action.new_insert_assignment:
            if not user.has_bcd(conn, insert_assignment.bcd_name):
                raise HTTPException(
                    status_code=status.HTTP_403_FORBIDDEN,
                    detail=f'You do not have permission to request patch for BCD {insert_assignment["bcd_name"]}'
                )

        internal_action.new_insert_assignment = action.new_insert_assignment
        internal_action.new_insert_type = action.new_insert_type

        p_port_old_dict = {}
        if action.p_port_gpk is not None:
            p_port_old_dict = {'gpk': action.p_port_gpk}

        r = execute_wapi_function(conn, [
            {'idx': 'module_list', 'name': 'nd.module.list', 'old': {'gpk': action.module_gpk}},
            {'idx': 'port_list', 'name': 'nd.p_port.list', 'old': p_port_old_dict,
                'inner_join_ref': {'module_list': 'default'}},
            {'idx': 'root_module_list', 'name': 'nd.module.list',
             'inner_join_ref': {'module_list': 'api_func_nd_module_hierarchy_gfk_list_is_subset'}},
            {'idx': 'room_list', 'name': 'nd.room.list', 'old': {'gpk': action.room_gpk},
             'inner_join_ref': {'root_module_list': 'default'}},
            {'idx': 'bldg_list', 'name': 'nd.bldg.list', 'old': {'gpk': action.building_gpk},
             'inner_join_ref': {'room_list': 'default'}},
            {'idx': 'site_list', 'name': 'nd.site.list', 'inner_join_ref': {'bldg_list': 'default'}},
            {
                'idx': 'tmp_all_ports',
                'name': 'nd.p_port.list',
                'inner_join_ref': {
                    'port_list': 'api_func_nd_p_port_is_contained_by_conn_id_gfk_list'
                },
                'old': {
                    'is_edge_node': True
                }
            },
            {
                'idx': 'inner_port',
                'name': 'nd.p_port.list',
                'inner_join_ref': {
                    'tmp_all_ports': 'self'
                },
                'anti_join_ref': {
                    'port_list': 'self'
                }
            },
            {
                'idx': 'inner_conn_port',
                'name': 'nd.p_port.list',
                'inner_join_ref': {
                    'inner_port': 'api_fkey_nd_p_port_conn_int'
                }
            },
            {
                'idx': 'inner_module',
                'name': 'nd.module.list',
                'inner_join_ref': {
                    'inner_port': 'default'
                }
            }
        ], superuser=True)
        if len(r['port_list']) != 1 and action.action != PatchActionType.CHANGE_INSERT:
            raise HTTPException(
                status_code=status.HTTP_404_NOT_FOUND,
                detail='The requested Port could not be found or chain invalid!'
            )
        if len(r['module_list']) != 1 or len(r['room_list']) != 1 or len(r['bldg_list']) != 1:
            raise HTTPException(
                status_code=status.HTTP_500_INTERNAL_SERVER_ERROR,
                detail='Ambiguous chain from building ({building}) to port ({port})! Giving up on that request!'.format(
                    building=action.building_gpk, port=action.p_port_gpk)
            )

        port = r['port_list'][0]
        if len(r['inner_port']) <= 0 or len(r['inner_module']) <= 0:
            # special case: port is a switch
            inner_port = port
            internal_is_switch = True
        else:
            inner_port = r['inner_port'][0]
            internal_is_switch = r['inner_module'][0]['is_active']

        inner_conn_port = None
        if internal_is_switch:
            tmp_inner_conn_ports = list(filter(lambda x: x.get('gpk', None) == inner_port['internal_connected_gfk'],
                                               r['inner_conn_port']))
            if len(tmp_inner_conn_ports) > 0:
                inner_conn_port = tmp_inner_conn_ports[0]

        port_name = inner_port['name'] if (internal_is_switch or inner_conn_port is None) else inner_conn_port['name'],

        internal_action.path_link = '{netdb_base}/~netadmin/netdoc/adm/cmp/cmp_mdlpas.cgi?ND_MDL.NAME={mdl_name}' \
                                    '&ND_P_PORT.NAME={port_name}&ND_BLDG.NR={bldg_nr}&ND_ROOM.NR={room_nr}' \
                                    '&MDL_PKG=netdoc&FRAMESET_PKG=netdoc&MDL_NAME=cmp_mdlpas&MDL_ACTION=out_trace' \
            .format(
                netdb_base=settings.netdb_admin_base,
                mdl_name=inner_port['mdl_fq_name'],
                port_name=port_name,
                bldg_nr=inner_port['mdl_bldg'],
                room_nr=inner_port['mdl_room']
            )

        internal_action.ports_summary = '{mdl_bldg} -> {mdl_room} -> {mdl_fq_name} ({mdl_type}) -> Port: {name} => ' \
                                        '{r_mdl_bldg} -> {r_mdl_room} -> {r_mdl_fq_name}::{r_name}' \
            .format(
                mdl_bldg=port['mdl_bldg'],
                mdl_room=port['mdl_room'],
                mdl_fq_name=port['mdl_fq_name'],
                name=port['name'],
                mdl_type=port['mdl_type'],
                r_mdl_bldg=inner_port['mdl_bldg'],
                r_mdl_room=inner_port['mdl_room'],
                r_mdl_fq_name=inner_port['mdl_fq_name'],
                r_name=inner_port['name'],
            )

        internal_action.p_port = Port(**r['port_list'][0])
        internal_action.module = Module(**r['module_list'][0])
        internal_action.room = Room(**r['room_list'][0])
        internal_action.building = Bldg(**r['bldg_list'][0])
        internal_action.site = Site(**r['site_list'][0])
        internal_action.action = action.action
        internal_action.bcd_tagged = action.bcd_tagged
        internal_action.comment = action.comment
        internal_actions.append(internal_action)

    return internal_actions


def render_jinja_template(path, template, **kwargs):
    template_loader = jinja2.FileSystemLoader(path)
    template_env = jinja2.Environment(loader=template_loader, autoescape=True)
    templ = template_env.get_template(template)
    return templ.render(**kwargs)


def send_informal_petch_request(message: str, mgr: Mgr, receiver=settings.patch_request_email_receivers,
                                reply_to: Optional[str] = None) -> bool:
    try:
        host_mode = db.host_omdl.OP_ENV_MODE.upper()

        path = pathlib.Path(__file__).parent.parent.resolve()
        TEMPLATE = 'templates/informal_patch_request_template.j2'

        body = render_jinja_template(path, TEMPLATE, message=message, mgr=mgr, host_mode=host_mode)

        subject = f'Informal Patch Request for {html.escape(mgr.first_name)} {html.escape(mgr.last_name)}'
        if host_mode:
            subject = f'[{host_mode}] {subject}'

        # hacky way to intercept mails for development. In production, the assignment should be used
        if settings.patch_request_email_receivers is not None:
            receiver = settings.patch_request_email_receivers

        send_email(to=receiver,
                   sender=settings.patch_request_email_sender,
                   reply_to=reply_to or mgr.email,
                   subject=subject,
                   body=body
                   )
        return True
    except Exception as e:
        raise e


def send_patch_request(actions, mgr: Mgr, receiver=settings.patch_request_email_receivers,
                       reply_to: Optional[str] = None) -> bool:
    try:
        host_mode = db.host_omdl.OP_ENV_MODE.upper()

        path = pathlib.Path(__file__).parent.parent.resolve()
        TEMPLATE = 'templates/patch_request_template.j2'

        body = render_jinja_template(path, TEMPLATE,
                                     actions=actions,
                                     mgr=mgr,
                                     host_mode=host_mode
                                     )

        building_numbers = ",".join(set([action.building.number for action in actions]))

        subject = f'Patch Request: ({html.escape(building_numbers)}) for {html.escape(mgr.first_name)} ' \
                  f'{html.escape(mgr.last_name)}'
        if host_mode:
            subject = f'[{host_mode}] {subject}'

        # hacky way to intercept mails for development. In production, the assignment should be used
        if settings.patch_request_email_receivers is not None:
            receiver = settings.patch_request_email_receivers

        send_email(to=receiver,
                   sender=settings.patch_request_email_sender,
                   reply_to=reply_to or mgr.email,
                   subject=subject,
                   body=body
                   )
        return True
    except Exception as e:
        raise e


def send_informal_patch_request_confirm(message: str, receiver) -> bool:
    try:
        host_mode = db.host_omdl.OP_ENV_MODE.upper()

        path = pathlib.Path(__file__).parent.parent.resolve()
        TEMPLATE = 'templates/informal_patch_request_confirm_template.j2'

        body = render_jinja_template(path, TEMPLATE,
                                     message=message,
                                     host_mode=host_mode
                                     )

        subject = 'Confirmation for Patch Request'
        if host_mode:
            subject = f'[{host_mode}] {subject}'

        # hacky way to intercept mails for development. In production, the assignment should be used
        if settings.patch_request_email_receivers is not None:
            receiver = settings.patch_request_email_receivers

        send_email(to=receiver,
                   sender=settings.patch_request_email_sender,
                   reply_to=settings.patch_request_email_sender,
                   subject=subject,
                   body=body
                   )
        return True
    except Exception as e:
        raise e


def send_patch_request_confirm(actions: list[PatchRequestAction], receiver) -> bool:
    try:
        host_mode = db.host_omdl.OP_ENV_MODE.upper()

        path = pathlib.Path(__file__).parent.parent.resolve()
        TEMPLATE = 'templates/patch_request_confirm_template.j2'

        body = render_jinja_template(path, TEMPLATE,
                                     actions=actions,
                                     host_mode=host_mode
                                     )

        building_numbers = ",".join(set([action.building.number for action in actions]))

        subject = f'Confirmation for Patch Request: ({html.escape(building_numbers)})'
        if host_mode:
            subject = f'[{host_mode}] {subject}'

        # hacky way to intercept mails for development. In production, the assignment should be used
        if settings.patch_request_email_receivers is not None:
            receiver = settings.patch_request_email_receivers

        send_email(to=receiver,
                   sender=settings.patch_request_email_sender,
                   reply_to=settings.patch_request_email_sender,
                   subject=subject,
                   body=body
                   )
        return True
    except Exception as e:
        raise e
